/* 	This program was inspired by the project MRBUS 
	developed by Nathan D. Holmes (maverick@drgw.net)
	(giving credit where credit is due), especially
	when it comes to the theory of the Revised 
	Differential Bus (see MRBus Concept Whitepaper)
	and its procedures like arbitration, packet 
	architecture and handling.
	See http://www.drgw.net/~maverick/electronics/mrbus              
	for more information
	
	This programm copyrighted 2004-2005 by 
	Michael Prader and Peter Stadler
	Licensed under the GNU GPL v2  
	See (when I get to it) http://www.b99.de.cx for more information */


// Version 0.9f, last revision 02/20/2005, known to be working, excluding the problems
// with the TX routine

#include "bus.h"


unsigned char ArbBitSend(unsigned char bit)
{
	char slice;
	char tmp=0;

	if (bit)
	{
		//TX is already 0 (mark)
		cbi(RS485PORT, RS485TXEN);
	} else {
		sbi(RS485PORT, RS485TXEN);
	}

	for (slice=0; slice<10; slice++)
	{
		if (slice > 2)
		{
			if (UARTPORT & UART_RXMASK) tmp=1;
			if (tmp ^ bit)
			{
				cbi(RS485PORT, RS485TXEN);
				sbi(SER_CONT_REG, RXEN);
				sbi(SER_CONT_REG, RXCIE);				
				sei();
				return(1);
			}
			
		}
		// delay_us(20);
		TCNT0 = 0x00;
		while (TCNT0 < __20_US);
	}
	return(0);
}


unsigned char xmit_packet(void)
{
	unsigned char status;
	unsigned char address;
	unsigned char i;
	uint16_t crc16_value;
	
	address = tx_buffer[PKT_SRC];
	
	if (state & TX_BUF_ACTIVE) return(1);
	
	// First calculate CRC16
	crc16_value=0x0000;
	
	for(i=0;i<tx_buffer[PKT_LEN];i++)
	{
		if ((i != PKT_CRC_H) && (i != PKT_CRC_L))
		{
			crc16_value = _crc16_update(crc16_value, tx_buffer[i]);
		}
	}
	tx_buffer[PKT_CRC_L] = (crc16_value & 0xFF);
	tx_buffer[PKT_CRC_H] = ((crc16_value >> 8) & 0xFF);	
		
	mrbus_activity=0;
	
	/* Begin of arbitration sequence */
	outp(0x00, TCNT1H);
	outp(0x00, TCNT1L);	
	while (TCNT1L < __2_MS);
	
	if (mrbus_activity) return(1);
	
	status = ((loneliness+priority)*10) + (address & 0x0F); 
	
	RS485PORT = 0x00;
	// Clear driver enable
	// set xmit pin low
	
	cbi(RS485DDR, RX); // Set as input
	
	cbi(SER_CONT_REG, TXEN);
	cbi(SER_CONT_REG, RXEN);  // Serial Port disable
	
	
	for (i=0; i<0x14; i++)
	{
		// delay_us(20);
		TCNT0 = 0x00;
		while(TCNT0 < __20_US);
		if (0 == (UARTPORT & UART_RXMASK))
		{
			sbi(SER_CONT_REG, RXEN);
			sbi(SER_CONT_REG, RXCIE);
			sei();
			return(1);
		}
	}

	/* Now, wait calculated time from above */
	for (i=0; i<status; i++)
	{
		// delay_us(10);
		TCNT0 = 0x00;
		while(TCNT0 < __10_US);
		if (0 == (UARTPORT & UART_RXMASK))
		{
			sbi(SER_CONT_REG, RXEN);
			sbi(SER_CONT_REG, RXCIE);
			sei();
			return(1);
		}
	}
	
	/* Arbitration Sequence - 4800 bps */
	/* Start Bit */
	if (ArbBitSend(0)) return(1);

	for (i=0; i<4; i++)
	{
		status = ArbBitSend(address & 0x01);
		address = address/2;

		if (status == 1) return(1);
	}	

	if (ArbBitSend(0)) return(1);
	
	for (i=0; i<4; i++)
	{
		status = ArbBitSend(address & 0x01);
		address = address/2;

		if (status == 1) return(1);
	}

	/* Stop Bit */
	if (ArbBitSend(1)) return(1);

	sbi(RS485PORT, TX);
	sbi(RS485PORT, RS485TXEN);
		
	tx_index=0;
	
	return(0); // Control over bus is assumed
}

SIGNAL(SIG_UART_RECV) //Receive Routine
{
	mrbus_activity = 1;
	if (SER_STAT_REG & RC_ERR_MASK)
	{
		// Handle framing errors - these are likely arbitration bytes
		rx_index = UDR;
		rx_index = 0; // Reset receive buffer
	}
	else
	{
		rx_input_buffer[rx_index++] = UDR;
		if ((rx_index > 5) && ((rx_index >= rx_input_buffer[PKT_LEN]) || (rx_index >= BUFFER_SIZE) ))
		{
			unsigned char i;
			rx_index = 0;
			if(!(state & RX_PKT_READY))
			{
				if (rx_input_buffer[PKT_LEN] > BUFFER_SIZE) rx_input_buffer[PKT_LEN] = BUFFER_SIZE;
				
				for(i=0; i < rx_input_buffer[PKT_LEN]; i++)
				{
					rx_buffer[i] = rx_input_buffer[i];
				}
				state = state | RX_PKT_READY;
				mrbus_activity = 2;
			}
		}
		if (rx_index >= BUFFER_SIZE)
		{
			rx_index = BUFFER_SIZE-1;
		}
	}
}

SIGNAL(SIG_UART_DATA) //Transmit Routine
{
	if ((tx_index >= BUFFER_SIZE) || (tx_index >= tx_buffer[PKT_LEN]))
	{	// Transmit is complete: terminate
		cbi(SER_CONT_REG, UDRIE);
		cbi(RS485PORT, RS485TXEN);
		sbi(SER_CONT_REG, RXEN);
		sbi(SER_CONT_REG, RXCIE);
		state = state & (~TX_BUF_ACTIVE);
		loneliness = 6;
	}
	else
	{
		UDR = tx_buffer[tx_index++];
		troppo=0x00;
		while (troppo < 0x02)
		{
			outp(0x00, TCNT0);	
			while (TCNT0 < 0xAF);
			troppo++;
		}
	}
}
