//#define AT90S8515
#define ATmega8515


#define F_CPU            16000000      /* 8Mhz */

#include <avr/io.h>
#include <avr/crc16.h>
#include <avr/eeprom.h>
#include <inttypes.h>
#include <avr/interrupt.h>
#include <avr/signal.h>



/********  INCLUDE HERE  *************/
#include "..\bus\bus.c"



/* EEPROM Configuration Addresses */
#define EE_DEVICE_ADDR    0x00

void init(void);
void PktHandler(void);

int main(void)
{
	unsigned char i;
	
	//start initialization
	init();
	
	i=0;
	while (i < BUFFER_SIZE) //initialize all buffers to 0x00
	{
		tx_buffer[i]=0x00;
		rx_buffer[i]=0x00;
		rx_input_buffer[i]=0x00;
		i++;
	}
	
	
	if (!(eeprom_is_ready())) eeprom_busy_wait();
	dev_addr = eeprom_read_byte(EE_DEVICE_ADDR);
	
	if (0xFF == dev_addr) {
		if (!(eeprom_is_ready())) eeprom_busy_wait();
		eeprom_write_byte(EE_DEVICE_ADDR, 0x00);
	}
	sei();
	
	PORTC = 0x03;
	
	troppo=0x00;
	while (troppo < 0x82) //ca 2 sek glaub ich
	{
		outp(0x00, TCNT1H);
		outp(0x00, TCNT1L);  
		while (TCNT1L < 0xFF);
		troppo++;
	}
	
	
	while(1)
	{
		if (state & RX_PKT_READY) PktHandler();
		
		while (state & TX_PKT_READY)
		{
			if (state & RX_PKT_READY) PktHandler();
			i = xmit_packet();
			if (i) //transmit start failed, wait and try again
			{
				
				outp(0x00, TCNT1H);
				outp(0x00, TCNT1L);	
				while (TCNT1L < __10_MS)
				{
					if (state & RX_PKT_READY) PktHandler();
				}
				loneliness--;
			}
			else
			{
				cli();
				state = state | TX_BUF_ACTIVE;
				SER_CONT_REG = (1<<RXEN) | (1<<TXEN);
				sbi(SER_CONT_REG, UDRIE);
				sei();
			}
			state = state & (~TX_PKT_READY);
		}
	}
}



void PktHandler(void)
{

	unsigned int addr;
	unsigned char i, crc_low, crc_high;
	unsigned short crc;
	crc = 0x0000;
	
	/************** PACKET FILTER *************/
	PORTC=0x04;
	/* Loopback Test - did we send it? */
	if (rx_buffer[PKT_SRC] == dev_addr) 
	{
		goto	PktExit;
	}
	PORTC=0x05;
	/* Destination Test - is this for us? */
	if ((rx_buffer[PKT_DEST] != 0xFF) && (rx_buffer[PKT_DEST] != dev_addr)) 
	{
		goto	PktExit;
	}
	PORTC=0x06;
	/* CRC16 Test - is the packet intact? */
	for(i=0; i < rx_buffer[PKT_LEN]; i++)
	{
		if ((i != PKT_CRC_H) && (i != PKT_CRC_L))
		{
			crc = _crc16_update(crc, rx_buffer[i]);
		}
	}
	
	crc_low = crc & 0xFF;
	crc_high = crc >> 8;
	
	if((crc_low != rx_buffer[PKT_CRC_L]) || (crc_high != rx_buffer[PKT_CRC_H])) goto PktExit;
	PORTC=0x07;

	/******** END PACKET FILTER *******/
	if ('A' == rx_buffer[PKT_TYPE])
	{
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 0x06;
		tx_buffer[PKT_TYPE] = 'a';
		state = state | TX_PKT_READY;
		goto PktExit;
	} else if ('R' == rx_buffer[PKT_TYPE])
	{
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 8;
		tx_buffer[PKT_TYPE] = 'r';
		tx_buffer[PKT_CV] = rx_buffer[PKT_CV];
		addr = rx_buffer[PKT_CV] & 0xFF;
		
		tx_buffer[PKT_VALUE] = eeprom_rb(addr);
		state = state | TX_PKT_READY;
		goto PktExit;		
	} else if ('W' == rx_buffer[PKT_TYPE]) {
		
		addr = rx_buffer[PKT_CV] & 0xFF;
		if (!eeprom_is_ready()) eeprom_busy_wait();
		eeprom_wb(addr, rx_buffer[PKT_VALUE]);
		tx_buffer[PKT_VALUE] = rx_buffer[PKT_VALUE];
		
		if (!eeprom_is_ready()) eeprom_busy_wait();
		dev_addr = eeprom_read_byte(EE_DEVICE_ADDR);
		
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 0x08;
		tx_buffer[PKT_TYPE] = 'w';
		tx_buffer[PKT_CV] = rx_buffer[PKT_CV];
		
		state = state | TX_PKT_READY;
		goto PktExit;
	} else if (0x14 == rx_buffer[PKT_TYPE]) {
		PORTC = rx_buffer[6];
		tx_buffer[PKT_SRC] = dev_addr;
		tx_buffer[PKT_DEST] = rx_buffer[PKT_SRC];
		tx_buffer[PKT_LEN] = 0x07;
		tx_buffer[PKT_TYPE] = 0x15;
		tx_buffer[6]=PORTC;
		state = state | TX_PKT_READY;
		goto PktExit;
	} 
	
	/*************** PACKET SUCCESS - PROCESS HERE ***************/
	
		
	PktExit:
	state = state & (~RX_PKT_READY);	
	return;
}

void init(void) //initializes the U(S)ART and the timer
{
	// set baud rate
	UBRRH=0x00;
	UBRR=UART_BAUD_SELECT;
	
	RS485DDR = (1<<RS485TXEN) | (1<<TX); //set TXen pin as output
	RS485PORT = 0x00;	// and low
	
	// Format: 8-N-1
	SER_CONT_REG = ((1<<TXEN) | (1<<RXEN) | (1<<RXCIE));
	UCSRC = (1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0);
	
	// prescaler is 8, timer counts 1/2 microseconds
	TCCR0 = 0x02;	
	/* prescaler is 1024 for 16bit timer */
	TCCR1A = 0x00;
	TCCR1B = 0x00 | (1<<CS12) | (1<<CS10);
	
	DDRC=0xFF;
	PORTC=0x01;
		
	ACSR = (1<<ACD);
	SFIOR=(1<<PUD);

}


